import os
import logging
import graphio
import json
import redis
import uuid
import operator
import time
import psutil
import jsonpickle
import traceback
from typing import Dict, Any, List, Type, Union, Generator

from dataclasses import dataclass, asdict

log = logging.getLogger(__name__)


class DrainNotReadyError(Exception):
    pass


@dataclass
class SetsMetaBase:
    """Metadata dataclass referencing a graphio.NodeSet or graphio.RelationshipSet in the cache."""

    # A key to reference the graphio set in the cache
    key: str
    # Size of the graphio set in the cache in bytes
    total_size_bytes: int
    # Indicator if the graphio set is a graphio.NodeSet or graphio.RelationshipSet
    type: Type = graphio.NodeSet

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self.key == other.key:
            return True
        return False

    def __hash__(self):
        return hash(self.key)


@dataclass
class NodeSetMeta(SetsMetaBase):
    labels: List[str] = None
    # unused. to be removed:
    # relations: List[dict] = None


@dataclass
class RelationSetMeta(SetsMetaBase):
    type: Type = graphio.RelationshipSet
    end_node_labels: List[str] = None
    start_node_labels: List[str] = None
    rel_type: str = None
    target_nodeSets: List[NodeSetMeta] = None


# ToDo: Document interface functions


class CacheInterface:
    """CacheInterface
    Abstract Class to describe the interface for the cache used by neobulkmp.
    The cache in neobulkmp is used for two tasks:

    * To buffer the workloads between WorkerSourcing (The processes parsing source data) and WorkerLoading (The processes loading sourced data into the database)
    * To pass through log messages from the workers to the manager process, which handle the logs centralistic

    The implemented cache class needs to have a high perfomance backend.
    Ideas for possible cache backends:

    * Redis DB - https://redis.io/ (✓ Implemented)
    * multiprocessing.shared_memory - https://docs.python.org/3/library/multiprocessing.shared_memory.html
    * Mongo DB - https://www.mongodb.com/de
    """

    def __init__(self, cache_backend_params: Dict = None, debug: bool = False):
        self.cache_backend_params = cache_backend_params
        self.debug = debug

    def test_connection(self) -> bool:
        """Test if the cache is ready to handle orders.

        Raises:
            NotImplementedError: [description]

        Returns:
            bool: True if cache is ready to play, False if cache is not available
        """
        raise NotImplementedError

    def clear_cache(self):
        """Clear any stuff in the cache from previous runs. this will we called when a new run is started"""
        raise NotImplementedError

    def storage_used(self) -> int:
        """Amount of KB the cache is using.

        Returns:
            int: KB of storage available
        """
        raise NotImplementedError

    def store_is_available(self) -> bool:
        """The backend can be temporary closed by the manager. Then sourcing worker are not allowed to store nodesets.
        This can make sense if, for example, the backend memory or storage is full. And the manager first needs to assign cores to loading workers to free up memory

        Returns:
            bool: [description]
        """
        raise NotImplementedError

    def close_store(self):
        """block the storage of NodeSets and RelationshipSets see self.store_is_available for more details

        Returns:
            bool: [description]
        """
        raise NotImplementedError

    def open_store(self):
        """Open a closed to store, to let sourcing worker continue their work

        Returns:
            bool: [description]
        """
        raise NotImplementedError

    def store_NodeSet(self, set: graphio.NodeSet):
        """Store the content of a graphio.NodeSet in the cache backend. this will be done by sourcing workers.

        Args:
            set (graphio.NodeSet): [description]
        """
        raise NotImplementedError

    def store_RelSet(self, set: graphio.RelationshipSet):
        """Store the content of a graphio.RelationshipSet in the cache backend. this will be done by sourcing workers.

        Args:
            set (graphio.RelationshipSet): [description]
        """
        raise NotImplementedError

    def fetch_NodeSets(self, nodeset_meta: SetsMetaBase) -> graphio.NodeSet:
        """Return a NodeSet of a certain type from the cache. It will be removed from the cache at the same time.
        The Nodeset type will be determined by the NodeSetMeta instance. A list of all available NodeSetMeta types can be received via 'list_SetsMeta()'
        Nodesets which have to be drained (because it was ordered by CacheInterface.order_RelSetDrain()) should get the highest priority when selecting the nodes to be returned

        Args:
            nodeset_meta (SetsMetaBase): An instance of SetsMetaBase (or its child NodeSetMeta) or  which describes the NodeSet type. A list of all available SetsMetaBase types can be received via 'list_SetsMeta()'

        Returns:
            graphio.NodeSet: [description]
        """
        raise NotImplementedError

    def fetch_RelSets(
        self, relset_meta: SetsMetaBase, drain_order_id: str
    ) -> graphio.RelationshipSet:
        """Return a RelationshipSet of a certain type from the cache. It will be removed from the cache at the same time.
        The RelationshipSet type will be determined by the RelationSetMeta instance. A list of all available RelationSet types can be received via 'list_SetsMeta()'

        In contrast to fetch_NodeSets() the caller need to order a so called 'drain' before fetching RelationshipSets.
        A drain, via order_RelSetDrain() will make sure that all current NodeSets that are connected to a type of RelationShipset are flushed to the Neo4j DB.
        A drain is done when is_drain_done() says so. After a drain is done, a loading worker is allowed call fetch_RelSets() and retrieve the RelatioshipSet sets that were in the cache during the drain order

        Args:
            relset_meta (RelationSetMeta): An instance of SetsMetaBase (or its child RelationSetMeta) or  which describes the NodeSet type. A list of all available SetsMetaBase types can be received via 'list_SetsMeta()'
            drain_order_id (str): The drain order id returned by order_RelSetDrain()

        Raises:
            DrainNotReadyError: When fetch_RelSets is called before the corresponding drain of related nodeSets is ready

        Returns:
            graphio.NodeSet: [description]
        """
        raise NotImplementedError

    def list_SetsMeta(
        self, set_type: Type[Union[graphio.NodeSet, graphio.RelationshipSet]] = None
    ) -> List[SetsMetaBase]:
        """returns a list of all RelationshipSet types and all NodesetSet types that are currently in the cache.
        The first item in the list, is item with the highest priority to be loaded in the database, the second item with second highest prio..and so on.
        This can be due to the size a Node- or Relationship-Set occupies in the Cache or due to a drain order

        Args:
            set_type (Type[Union[graphio.NodeSet, graphio.RelationshipSet]], optional): [description]. Defaults to None. - Filter for NodeSets or Relationship types only

        Returns:
            List[SetsMetaBase]: List of graph sets description order by priority
        """
        raise NotImplementedError

    def get_NodeSetMeta(self, nodeset_meta: SetsMetaBase) -> NodeSetMeta:
        """Returns a more detailed description of a NodeSet (compared to SetsMetaBase), like labels and relations. NodeSetMeta is a child class of SetsMetaBase

        Args:
            nodeset_meta (SetsMetaBase): [description]

        Returns:
            NodeSetMeta: [description]
        """
        raise NotImplementedError

    def get_RelSetMeta(self, relset_meta: SetsMetaBase) -> RelationSetMeta:
        """Returns a more detailed description of a RelationSet(compared to SetsMetaBase), like end_node_labels,start_node_labels,rel_type and target nodeSets types. NodeSetMeta is a child class of SetsMetaBase

        Args:
            relset_meta (SetsMetaBase): [description]

        Raises:
            NotImplementedError: [description]

        Returns:
            RelationSetMeta: [description]
        """
        raise NotImplementedError

    def order_RelSetDrain(self, relset_meta: SetsMetaBase) -> str:
        """
        A drain order on a NodeSet insures the caller that all nodeset that are currently in the cache are empty as soon as 'CacheInterface.is_drain_done' returns True
        This is needed for the database loading workers when loading RelationSets. Workers then can be certain that all nodes that could be attached to the RelationSet are now in the DB

        Args:
            relset_meta (relset_meta): Meta data to a cached nodesets. Needed to identify drain slot. Can be obtain via 'CacheInterface.list_SetsMeta'

        Returns:
            str: drain order ticket. identifies the drain order. Needed to ask if drain order is done via 'CacheInterface.is_drain_done'
        """
        raise NotImplementedError

    def report_SetLoaded(
        self, nodeset: Union[graphio.NodeSet, graphio.RelationshipSet]
    ):
        """A loading worker will report here when it just finished to loading a graphio.NodeSet or graphio.RelationshipSet into the neo4j database.
        This information can be used to update the status of drain orders and to count the nodes/rels loaded. the amount of loaded nodes/rels can be queried via 'get_report_graphsets_loaded'

        Args:
            nodeset (Union[graphio.NodeSet, graphio.RelationshipSet]): [description]
        """
        raise NotImplementedError

    def is_drain_done(self, drain_ticket: str) -> bool:
        raise NotImplementedError

    def get_report_graphsets_loaded(self) -> Dict[str, List[Dict[str, int]]]:
        """Returns a list of NodeSet types and RelationshipSet types with the amount of nodes allready loaded

        Returns:
            List[Dict]: List LabelCombinations loaded per NodeSet worker with the nodes allready loaded e.g. '{"Nodes":[{"MyLabel:OtherLabel":23421},{"OneLabel":3846}],"Relations":[{"LABEL_HAS_OTHERLABEL":234}]}'
        """
        raise NotImplementedError

    def write_log(self, record: logging.LogRecord):
        raise NotImplementedError

    def fetch_logs(self) -> Generator[logging.LogRecord, None, None]:
        raise NotImplementedError

    def get_logs(self) -> List[logging.LogRecord]:
        raise NotImplementedError


# ToDo: Refactor RedisCache
# * evaluate to mount drain_orders to set meta data instead of workers
# * Find other opportunities to simplifiy redis cache code


class RedisCache(CacheInterface):
    NODESET_PREFIX = "NODESET:"
    RELSET_PREFIX = "RELSET:"
    DRAIN_ORDERS_KEY = "DRAIN_ORDERS_OPEN"
    DRAIN_ORDERS_READY_KEY = "DRAIN_ORDERS_READY"
    DRAIN_ORDERS_NODESETS_DONE_KEY = "DRAIN_ORDERS_NODESETS_DONE"

    # Fetch COMBINE_NODESET_COUNT amount of nodesets and combine to one for dataloading
    COMBINE_SET_COUNT = 100

    @dataclass
    class DrainOrder:
        id: str
        relset_meta: RelationSetMeta
        nodesets_meta: List[NodeSetMeta]

        buckets_key_base: str
        relset_bucket_key: str
        nodesets_buckets_keys: List[str]
        time_of_drain_order: float
        preceding_overlapping_drain_orders: List["RedisCache.DrainOrder"]

        @classmethod
        def create_order(
            cls, relset_meta: RelationSetMeta, cache: "RedisCache"
        ) -> "RedisCache.DrainOrder":
            drain_order_id = str(uuid.uuid4())
            drain_timestamp = time.time()
            buckets_key_base = f"DRAIN:{drain_order_id}:{drain_timestamp}"
            # refresh meta data
            relset_meta = cache.get_RelSetMeta(relset_meta)
            drain_order = cls(
                id=drain_order_id,
                relset_meta=relset_meta,
                nodesets_meta=relset_meta.target_nodeSets,
                buckets_key_base=buckets_key_base,
                relset_bucket_key=[],
                nodesets_buckets_keys=[],
                time_of_drain_order=drain_timestamp,
                preceding_overlapping_drain_orders=[],
            )

            return drain_order._materialize_drain_order(cache)

        @classmethod
        def get_by_order_id(
            cls, drain_order_id: uuid.UUID, cache: "RedisCache"
        ) -> "RedisCache.DrainOrder":
            p = cache.redis.pipeline()
            p.hget(cache.DRAIN_ORDERS_KEY, drain_order_id)
            p.hget(cache.DRAIN_ORDERS_READY_KEY, drain_order_id)
            results = p.execute()
            for res in results:
                if res:
                    return jsonpickle.loads(res)

        @classmethod
        def list_orders(
            cls, cache, status=None, nodeset_meta_filter: NodeSetMeta = None
        ) -> List["RedisCache.DrainOrder"]:
            p = cache.redis.pipeline()
            if status in [None, "OPEN"]:
                p.hgetall(cache.DRAIN_ORDERS_KEY)
            if status in [None, "READY"]:
                p.hgetall(cache.DRAIN_ORDERS_READY_KEY)
            results = p.execute()
            orders: List[RedisCache.DrainOrder] = []
            for res in results:
                for drain_id, raw_drain_order in res.items():
                    order: RedisCache.DrainOrder = jsonpickle.loads(raw_drain_order)
                    # If a nodeset filter is set, we search for every drain order thats affecting the nodeset in nodeset_meta_filter
                    if nodeset_meta_filter is None or (
                        nodeset_meta_filter is not None
                        and not set(nodeset_meta_filter.labels).isdisjoint(
                            set(
                                order.relset_meta.end_node_labels
                                + order.relset_meta.start_node_labels
                            )
                        )  # nodeset_meta_filter.labels in [ns.key for ns in order.nodesets_meta]
                    ):
                        orders.append(order)
            # Sort by oldest drain order first.
            return sorted(orders, key=operator.attrgetter("time_of_drain_order"))

        @classmethod
        def list_open_nodesets_meta(cls, cache: "RedisCache") -> List[NodeSetMeta]:
            """Return a list of all NodeSetMeta that are affected by a drain
            Args:
                cache (RedisCache): [description]

            Returns:
                List[NodeSetMeta]: [description]
            """
            open_nodesets_meta = []

            # Get a list of all allready drained buckets keys that are part of an open drain order
            all_drained_buckets_keys = [
                bk.decode("utf-8")
                for bk in cache.redis.lrange(
                    cache.DRAIN_ORDERS_NODESETS_DONE_KEY, 0, -1
                )
            ]
            # Get a list of all open drain orders
            all_open_drain_orders = cls.list_orders(cache, "OPEN")
            for drain_order in all_open_drain_orders:
                for drain_nodeset_meta, bucket_key in zip(
                    drain_order.nodesets_meta, drain_order.nodesets_buckets_keys
                ):
                    if (
                        bucket_key
                        not in all_drained_buckets_keys
                        + ["DRAIN_EMPTY_PLACEHOLDER_KEY"]
                        and drain_nodeset_meta not in open_nodesets_meta
                    ):
                        open_nodesets_meta.append(drain_nodeset_meta)
            return list(reversed(open_nodesets_meta))

        @classmethod
        def get_drain_bucket_keys(
            cls,
            cache: "RedisCache",
            graphset_meta: SetsMetaBase = None,
            set_type: Type[Union[graphio.NodeSet, graphio.RelationshipSet]] = None,
        ):
            bucket_keys = []
            all_open_drain_orders = cls.list_orders(cache, "OPEN")
            for drain_order in all_open_drain_orders:

                if set_type in [None, graphio.NodeSet]:
                    for bucket_key in drain_order.nodesets_buckets_keys:
                        if bucket_key != "DRAIN_EMPTY_PLACEHOLDER_KEY":
                            if graphset_meta is None or (
                                graphset_meta is not None
                                and graphset_meta in drain_order.nodesets_meta
                            ):
                                bucket_keys.append(bucket_key)
                if set_type in [None, graphio.RelationshipSet]:
                    if graphset_meta is None or (
                        graphset_meta is not None
                        and graphset_meta == drain_order.relset_meta
                    ):
                        bucket_keys.append(drain_order.relset_bucket_key)
            return bucket_keys

        def is_empty(self, cache):
            # if a drain order has no buckets (only "DRAIN_EMPTY_PLACEHOLDER_KEY" placeholders), it will never recieve a set_loaded message and can go stale
            all_drained_buckets_keys = [
                bk.decode("utf-8")
                for bk in cache.redis.lrange(
                    cache.DRAIN_ORDERS_NODESETS_DONE_KEY, 0, -1
                )
            ]
            for bucket in self.nodesets_buckets_keys:
                if bucket != "DRAIN_EMPTY_PLACEHOLDER_KEY" and (
                    bucket not in all_drained_buckets_keys
                    or cache.redis.llen(bucket) > 0
                ):
                    return False
            for preceding_order in self.preceding_overlapping_drain_orders:
                if preceding_order.get_status(cache) == "OPEN":
                    return False
            return True

        def get_status(self, cache):
            if self.is_empty(cache):
                self._set_status(cache, "READY")
                return "READY"
            p = cache.redis.pipeline()
            p.hget(cache.DRAIN_ORDERS_KEY, self.id)
            p.hget(cache.DRAIN_ORDERS_READY_KEY, self.id)
            results = p.execute()
            if results[0]:
                return "OPEN"
            if results[1]:
                return "READY"
            return "FINISHED"

        def _set_status(self, cache: "RedisCache", status: str):
            if status == "READY":

                p = cache.redis.pipeline()
                p.multi()
                p.hdel(cache.DRAIN_ORDERS_KEY, self.id)
                p.hset(cache.DRAIN_ORDERS_READY_KEY, self.id, jsonpickle.dumps(self))
                p.execute()
            elif status == "FINISHED":
                # Delete drain order
                p = cache.redis.pipeline()
                p.multi()
                p.hdel(cache.DRAIN_ORDERS_READY_KEY, self.id)
                p.execute()
                # Delete related nodeset-done entries
                to_be_deleted_entries = []
                for key in cache.redis.lrange(
                    cache.DRAIN_ORDERS_NODESETS_DONE_KEY, 0, -1
                ):
                    key = key.decode()
                    if self.id in key:
                        to_be_deleted_entries.append(key)
                for key in to_be_deleted_entries:
                    cache.redis.lrem(cache.DRAIN_ORDERS_NODESETS_DONE_KEY, 0, key)

            else:
                raise ValueError(f"Expected 'READY' or 'FINISHED' got {status}")

        def get_nodeset(self, cache: "RedisCache", nodeset_meta) -> graphio.NodeSet:
            """Return nodeset from cache drain bucket

            Args:
                cache (RedisCache): [description]
                nodeset_meta ([type]): [description]

            Returns:
                graphio.NodeSet: [description]
            """

            for bucket_key, drain_nodeset_meta in zip(
                self.nodesets_buckets_keys, self.nodesets_meta
            ):
                if nodeset_meta.key == drain_nodeset_meta.key:
                    if bucket_key != "DRAIN_EMPTY_PLACEHOLDER_KEY":
                        nodeSet_parameter_dict = cache._parse_key(
                            nodeset_meta.key, cache.NODESET_PREFIX
                        )
                        ns = graphio.NodeSet.from_dict(
                            {
                                **nodeSet_parameter_dict,
                                **{
                                    "nodes": cache._fetch_list_range(
                                        bucket_key, cache.COMBINE_SET_COUNT, True
                                    )
                                },
                            }
                        )
                        ns.drain_order = self
                        ns.drain_bucket_key = bucket_key
                        ns.nodeset_meta = nodeset_meta
                        return ns
                    else:
                        return None

        def get_relset(self, cache: "RedisCache") -> graphio.RelationshipSet:
            """Return RelationshipSet from cache

            Args:
                cache (RedisCache): [description]
                nodeset_meta ([type]): [description]

            Returns:
                graphio.NodeSet: [description]
            """
            relSet_parameter_dict = cache._parse_key(
                self.relset_meta.key, cache.RELSET_PREFIX
            )
            rs = graphio.RelationshipSet.from_dict(
                {
                    **relSet_parameter_dict,
                    **{
                        "relationships": cache._fetch_list_range(
                            self.relset_bucket_key, None, True
                        )
                    },
                }
            )
            rs.drain_order = self
            rs.drain_bucket_key = self.relset_bucket_key
            rs.nodeset_meta = self.relset_meta
            return rs

        def _set_nodeset_status_drained(
            self, cache: "RedisCache", nodeset: graphio.NodeSet
        ):
            if nodeset.drain_order.id == self.id:
                # is bucket not a placeholder and is bucket not empty:
                if (
                    nodeset.drain_bucket_key != "DRAIN_EMPTY_PLACEHOLDER_KEY"
                    and cache.redis.llen(nodeset.drain_bucket_key) == 0
                ):
                    log.debug(f"SET nodeset {nodeset.labels} for {self.id} as drained")
                    cache.redis.rpush(
                        cache.DRAIN_ORDERS_NODESETS_DONE_KEY, nodeset.drain_bucket_key
                    )
                    return True
                else:
                    # Bucket is not empty. we wont set status to drained
                    return False
            else:
                raise ValueError(
                    f"NodeSet '{':'.join(nodeset.labels)}' is not target of DrainOrder:id'{self.id}'"
                )

        def are_nodesets_drained(self, cache: "RedisCache"):
            # Check preceding overlapping drain orders
            for other_drain_orders in self.preceding_overlapping_drain_orders:
                if not other_drain_orders.are_nodesets_drained(cache):
                    return False
            # Check self

            # check first if drain order is still existent
            if self.get_status(cache) == "FINISHED":
                return True

            # Get a list a fullfilled drained nodeset buckets
            all_drained_buckets_keys = [
                bucketkey.decode("utf-8")
                for bucketkey in cache.redis.lrange(
                    cache.DRAIN_ORDERS_NODESETS_DONE_KEY, 0, -1
                )
            ]

            # get own drain order buckets
            self_bucket_keys = [
                bucketkey
                for bucketkey in self.nodesets_buckets_keys
                if bucketkey != "DRAIN_EMPTY_PLACEHOLDER_KEY"
            ]
            # check if all own buckets appear in fullfilled bucket list
            if set(self_bucket_keys).issubset(set(all_drained_buckets_keys)):
                return True
            else:
                return False

        @classmethod
        def set_graphset_drained(
            cls,
            cache: "RedisCache",
            graphset: Union[graphio.NodeSet, graphio.RelationshipSet],
        ):
            if isinstance(graphset, graphio.NodeSet):
                if cache.debug:
                    all_drained_buckets_keys = [
                        bucketkey.decode("utf-8")
                        for bucketkey in cache.redis.lrange(
                            cache.DRAIN_ORDERS_NODESETS_DONE_KEY, 0, -1
                        )
                    ]
                    if graphset.drain_bucket_key in all_drained_buckets_keys:
                        raise ValueError(
                            f"Graphset({len(graphset.nodes)} nodes) '{graphset.drain_bucket_key}' is allready set as drained. It was called a second time :("
                        )

                graphset.drain_order._set_nodeset_status_drained(cache, graphset)
                if graphset.drain_order.are_nodesets_drained(cache):
                    log.debug(
                        f"Drain order for '{graphset.drain_order.relset_meta.rel_type}' is ready. RelationSet can be loaded now"
                    )
                    graphset.drain_order._set_status(cache, "READY")
            elif isinstance(graphset, graphio.RelationshipSet):
                log.debug(f"Drain order for '{graphset.rel_type}' is finished")
                graphset.drain_order._set_status(cache, "FINISHED")

        def set_drain_order_done(self, cache):
            p = cache.redis.pipeline()
            p.hdel(cache.DRAIN_ORDERS_KEY, self.id)
            p.lpush("DRAINS_DONE", self.id)
            p.execute()

        def _materialize_drain_order(self, cache):
            """
            To actually create the drain order we need to
                * seperate the relSet all affected nodesets each to a "drain bucket"
                * get all other current drain orders to later check if there are overlapping drain orders (relationshipSets that are targeting at the same nodeset)
                * make an entry in the redis drain order list to proclaim the drain order (for other processes to interact with the drain oder)
            """
            # Get other drain orders, to check overlaps
            other_drain_orders = cache.redis.hgetall(cache.DRAIN_ORDERS_KEY)

            # Check other drain order for overlap and update drain order if there are other drainorders
            overlapping_drain_orders = []
            for drain_id, pickled_drain_order in other_drain_orders.items():
                other_drain_order: RedisCache.DrainOrder = jsonpickle.loads(
                    pickled_drain_order
                )
                # if other drain order shares one or two nodesets its overlapping
                if [
                    other_ns
                    for other_ns in other_drain_order.nodesets_meta
                    if other_ns in self.nodesets_meta
                ]:
                    overlapping_drain_orders.append(other_drain_order)

            if overlapping_drain_orders:
                # update drain order if there were overlaping other drain orders
                self.preceding_overlapping_drain_orders = overlapping_drain_orders

            # Move nodesets to drain buckets
            self.nodesets_buckets_keys = []
            for set_meta in self.nodesets_meta:
                drain_key = f"{self.buckets_key_base}:{set_meta.key}"
                # Move current content of set to a new location
                try:
                    cache.redis.rename(set_meta.key, drain_key)
                except redis.exceptions.ResponseError:
                    # To be drained nodeset was empty not existent.
                    # atm we can not rename it.
                    # this happens when a beforehand worker just ordered a drain or the nodeset was coincidentally just loaded
                    # We will just provide a placeholder key
                    drain_key = "DRAIN_EMPTY_PLACEHOLDER_KEY"
                self.nodesets_buckets_keys.append(drain_key)
            # generate drain bucket for relatioshipset
            self.relset_bucket_key = f"{self.buckets_key_base}:{self.relset_meta.key}"
            # move relationset to drain bucket
            try:
                cache.redis.rename(self.relset_meta.key, self.relset_bucket_key)
            except redis.exceptions.ResponseError:
                log.error(
                    f"Failed renaming {self.relset_meta.key} to {self.relset_bucket_key}"
                )
                raise
            # proclaim the drain order: Save the meta data of the drain a key value pair a redis hashlist called cache.DRAIN_ORDERS_KEY

            cache.redis.hmset(cache.DRAIN_ORDERS_KEY, {self.id: jsonpickle.dumps(self)})

            # Check if nodesets are empty and drain is allready fullfilled
            if self.are_nodesets_drained(cache):
                self._set_status(cache, "READY")
            return self

    def __init__(self, cache_backend_params: Dict = None, debug: bool = False):
        super(type(self), self).__init__()
        if cache_backend_params is None:
            cache_backend_params = {"host": "localhost"}
        self.redis = redis.Redis(**cache_backend_params)

    def clear_cache(self):
        self.redis.delete("LOAD_REPORT")
        for key in self.redis.keys("RELSET*"):
            self.redis.delete(key)
        for key in self.redis.keys("NODESET*"):
            self.redis.delete(key)
        for key in self.redis.keys("DRAIN*"):
            self.redis.delete(key)

    def test_connection(self) -> bool:
        res = self.redis.info()
        if res:
            return True
        else:
            return False

    def storage_used(self) -> int:
        return self.redis.execute_command("INFO")["used_memory"]

    def store_is_available(self) -> bool:
        if self.redis.get("STORE_STATUS"):
            # Store is closed by manager
            return False
        else:
            return True

    def close_store(self) -> bool:
        """block the storage of NodeSets and RelationshipSets see self.store_is_available for more details

        Returns:
            bool: [description]
        """
        self.redis.set("STORE_STATUS", "CLOSED")

    def open_store(self) -> bool:
        """Open a closed to store, to let sourcing worker continue their work

        Returns:
            bool: [description]
        """
        self.redis.delete("STORE_STATUS")

    def store_NodeSet(self, set: graphio.NodeSet):
        set_dict = set.to_dict()
        data = set_dict.pop("nodes", [])
        cache_key = self.NODESET_PREFIX + json.dumps(set_dict)
        self.redis.lpush(cache_key, json.dumps(data))
        log.debug(f"store_NodeSet '{cache_key}'")
        # self.redis.lpush(cache_key, json.dumps(data))

    def store_RelSet(self, set: graphio.RelationshipSet):
        data = {}
        set_dict = set.to_dict()
        data = set_dict.pop("relationships", [])
        cache_key = self.RELSET_PREFIX + json.dumps(set_dict)
        self.redis.lpush(cache_key, json.dumps(data))
        log.debug(f"store_RelSet '{cache_key}'")

    def _fetch_list_range(
        self, key: str, range_size: int = None, parse_and_combine=True
    ) -> List:
        # fetch and delete from cache in one transaction
        p = self.redis.pipeline()
        # watch key for change
        # p.watch(key)
        # starts transactional block of pipeline
        p.multi()

        if range_size is None:
            # get all nodesets
            p.lrange(key, 0, -1)
            # delete key
            p.delete(key)
        else:
            # get first 'range_size' nodesets
            p.lrange(key, 0, range_size - 1)
            # delete first 'range_size' nodesets
            p.ltrim(key, range_size, -1)

        # ends transactional block of pipeline
        result = p.execute()
        if parse_and_combine:
            return [item for sublist in result[0] for item in json.loads(sublist)]
        else:
            return result[0]

    def _parse_key(self, key: str, prefix: str) -> Dict:
        key_data = None
        if isinstance(key, bytes):
            key_str = key.decode()
        else:
            key_str = key

        if key_str.startswith(prefix):
            key_data = json.loads(key[len(prefix) :])
        else:
            raise ValueError(
                f"Wrong key type. Expected '{prefix}{{somekeydata}}. Got {key}"
            )
        return key_data

    def fetch_NodeSets(self, nodeset_meta: NodeSetMeta) -> graphio.NodeSet:
        # Decode keydata for creating ndoeset
        # example keydata: 'NODESET:{"labels": ["Movie"], "merge_keys": "title"}'

        # Check if there are any open drainOrders on this type of NodeSet
        # If yes we return the oldest/first drain bucket of this nodeset
        if isinstance(nodeset_meta, SetsMetaBase):
            nodeset_meta = self.get_NodeSetMeta(nodeset_meta)
        drain_orders = RedisCache.DrainOrder.list_orders(
            self, status="OPEN", nodeset_meta_filter=nodeset_meta
        )
        if drain_orders:
            ns = drain_orders[0].get_nodeset(self, nodeset_meta=nodeset_meta)
            if ns:
                return ns

        # there are no drain orders, we can return content from the general bucket
        nodeSet_parameter_dict = self._parse_key(nodeset_meta.key, self.NODESET_PREFIX)
        # Fetch data from cache, combine result of different workers to one large nodeset and create a graphio.NodeSet based on that data
        return graphio.NodeSet.from_dict(
            {
                **nodeSet_parameter_dict,
                **{
                    "nodes": self._fetch_list_range(
                        nodeset_meta.key, self.COMBINE_SET_COUNT, True
                    )
                },
            }
        )

    def fetch_RelSets(
        self, relset_meta: SetsMetaBase, drain_order_id: str
    ) -> graphio.RelationshipSet:
        # Decode key data
        # example key: "RELSET:{\"rel_type\": \"ACTS_IN\", \"start_node_labels\": [\"Actor\"], \"end_node_labels\": [\"Movie\"], \"start_node_properties\": [\"name\"], \"end_node_properties\": [\"title\"], \"unique\": false}"
        drain_order = RedisCache.DrainOrder.get_by_order_id(drain_order_id, self)
        if self.debug and not drain_order.are_nodesets_drained(self):
            log.exception(
                f"Drain not finished, but relationship was tried to fetch. drain order id {drain_order_id}"
            )
            raise DrainNotReadyError

        relset_meta_dict = self._parse_key(relset_meta.key, self.RELSET_PREFIX)
        # Fetch relatipshipSet from bucket
        return drain_order.get_relset(self)

    def list_SetsMeta(
        self, set_type: Type[Union[graphio.NodeSet, graphio.RelationshipSet]] = None
    ) -> List[SetsMetaBase]:
        meta_data = []
        if set_type in [None, graphio.NodeSet]:
            # fetch all keys begining with NODESET_PREFIX and pack each into an SetsMetaBase instance
            set_keys = {
                key.decode("utf-8")
                for key in self.redis.keys(f"{self.NODESET_PREFIX}*")
            }
            """drain_nodeset_keys = {
                key.decode("utf-8").split(":", 3)[3]
                for key in self.redis.keys(f"DRAIN:*:{self.NODESET_PREFIX}*")
            }
            nodeset_keys = nodeset_keys | drain_nodeset_keys
            log.info(f"nodeset_keys: {nodeset_keys}")"""
            for nkey in set_keys:
                meta_entry = SetsMetaBase(
                    key=nkey, total_size_bytes=0, type=graphio.NodeSet
                )
                meta_entry.total_size_bytes = self._get_NodeSet_real_size(meta_entry)
                meta_data.append(meta_entry)

        if set_type in [None, graphio.RelationshipSet]:
            # fetch all keys begining with RELSET_PREFIX and pack each into an SetsMetaBase instance
            set_keys = [
                key.decode("utf-8") for key in self.redis.keys(f"{self.RELSET_PREFIX}*")
            ]
            meta_data = []
            for key in set_keys:
                # TODO: maybe it would make sense here to add related NodeSet size into total_size_bytes? this would prioritize larger nodesets to be drained first. does this make sense? think about it!
                meta_entry = SetsMetaBase(
                    key=key,
                    total_size_bytes=self.redis.memory_usage(key),
                    type=graphio.RelationshipSet,
                )
                meta_data.append(meta_entry)
        # Sort by priority (Drainorders first, then by size)
        meta_data.sort(key=operator.attrgetter("total_size_bytes"), reverse=True)

        if set_type in [None, graphio.NodeSet]:
            # Move nodesets with drain order to the front
            for open_drain_nodeset_meta in list(
                reversed(RedisCache.DrainOrder.list_open_nodesets_meta(cache=self))
            ):
                # check if nodeset appears in list of nodeset with a drainOrder, if yes get index this nodeset in meta_data list
                index = next(
                    (
                        i
                        for i, md in enumerate(meta_data)
                        if md.key in open_drain_nodeset_meta.key
                    ),
                    None,
                )
                if index not in [None, 0]:
                    # if nodeset_meta appeard in a drain order move it to the front
                    # pop it first into tmp var...
                    md = meta_data.pop(index)
                elif index is None:
                    # if drain order references a nodeset that was not fetched into 'meta_data'
                    # this can happen when a drain order justed moved the whole nodeset content to a drain bucket and no new nodeset were supplied by sourcing workers
                    md = open_drain_nodeset_meta
                    md.total_size_bytes = self._get_NodeSet_real_size(md)
                if index != 0:
                    # ... insert it to the front of the list
                    meta_data.insert(0, md)
        return meta_data

    def get_NodeSetMeta(self, nodeset_meta: SetsMetaBase) -> NodeSetMeta:
        node_meta = NodeSetMeta(
            key=nodeset_meta.key,
            total_size_bytes=nodeset_meta.total_size_bytes,  # self._get_NodeSet_real_size(nodeset_meta) # refresh ob total size is not used/needed so far. we can shave that off
        )
        node_meta.labels = self._parse_key(nodeset_meta.key, self.NODESET_PREFIX)[
            "labels"
        ]
        return node_meta

    def _get_NodeSet_real_size(self, nodeSet_meta: SetsMetaBase) -> int:
        # NodeSets of one key can be distributed in many drain buckets
        # to get the real site of a nodeset in chache we need to add alle these buckets together with the origin list

        bucket_keys = RedisCache.DrainOrder.get_drain_bucket_keys(
            cache=self, graphset_meta=nodeSet_meta, set_type=graphio.NodeSet
        )
        size = 0
        for bucket_key in bucket_keys + [nodeSet_meta.key]:
            dsize = self.redis.memory_usage(bucket_key)
            if dsize:
                size += dsize
        return size

    def get_RelSetMeta(self, relset_meta: SetsMetaBase) -> RelationSetMeta:
        rel_meta = RelationSetMeta(
            key=relset_meta.key,
            total_size_bytes=self.redis.memory_usage(relset_meta.key),
        )
        rel_key_data = self._parse_key(rel_meta.key, self.RELSET_PREFIX)
        rel_meta.rel_type = rel_key_data["rel_type"]
        rel_meta.start_node_labels = rel_key_data["start_node_labels"]
        rel_meta.end_node_labels = rel_key_data["end_node_labels"]
        rel_meta.target_nodeSets = []
        all_node_sets = self.list_SetsMeta(set_type=graphio.NodeSet)
        for nodeset in all_node_sets:
            nodeset_data = self._parse_key(nodeset.key, self.NODESET_PREFIX)
            if set(nodeset_data["labels"]) in [
                set(rel_key_data["start_node_labels"]),
                set(rel_key_data["end_node_labels"]),
            ]:
                rel_meta.target_nodeSets.append(nodeset)
        return rel_meta

    def order_RelSetDrain(self, relset_meta: SetsMetaBase) -> str:

        drain_order: self.DrainOrder = self.DrainOrder.create_order(relset_meta, self)
        log.debug(f"Order Drain for {drain_order.relset_meta.rel_type}")
        return drain_order.id

    def is_drain_done(self, drain_order: Union["RedisCache.DrainOrder", str]) -> bool:
        d_order = None
        if isinstance(drain_order, str):
            d_order = RedisCache.DrainOrder.get_by_order_id(drain_order, self)
        elif isinstance(drain_order, RedisCache.DrainOrder):
            d_order = drain_order
        else:
            raise ValueError(
                f"Expected 'RedisCache.DrainOrder' or 'str', got {type(drain_order)}"
            )
        if d_order.get_status(self) in ["READY"]:
            return True
        else:
            return False

    def report_SetLoaded(
        self, graphset: Union[graphio.NodeSet, graphio.RelationshipSet]
    ):
        if type(graphset) == graphio.NodeSet:
            key = f"NODESET:{':'.join(graphset.labels)}"
            amount = len(graphset.nodes)
        elif type(graphset) == graphio.RelationshipSet:
            key = f"RELSET:{graphset.rel_type}"
            amount = len(graphset.relationships)
        self.redis.hincrby("LOAD_REPORT", key, amount)
        if hasattr(graphset, "drain_order"):
            RedisCache.DrainOrder.set_graphset_drained(self, graphset)

    def get_report_graphsets_loaded(self) -> Dict[str, List[Dict[str, int]]]:
        result = {"Nodes": {}, "Relations": {}}
        for key, amount in self.redis.hgetall("LOAD_REPORT").items():
            key = key.decode("utf-8")
            amount = int(amount)
            if key.startswith("NODESET:"):
                result["Nodes"][key[len("NODESET:") :]] = amount
            elif key.startswith("RELSET:"):
                result["Relations"][key[len("RELSET:") :]] = amount
        return result

    def write_log(self, record: logging.LogRecord):
        # self.redis.lpush("LOGSTEST", jsonpickle.dumps(record))
        try:
            self.redis.rpush("LOG_LIST", jsonpickle.dumps(record))
            # pusblish seems to be broken; see fetch_logs() comments.
            # self.redis.publish("LOG_LIST", jsonpickle.dumps(record)))
        except:
            print(traceback.format_exc())
            self.redis.rpush("LOG_LIST", traceback.format_exc())
            raise

    def fetch_logs(self) -> Generator[logging.LogRecord, None, None]:
        """redis-py pub/sub is broken somehow. always only returns '1' (as int) as data

        pub = self.redis.pubsub()
        pub.subscribe("LOG_STREAM")
        for raw_record in pub.listen():
            yield jsonpickle.loads(raw_record)
        """

        # for now we will use a plain redis list as a queue for the time being (instead of the broken pub/sub system).
        # in a later iteration i would like to switch from redis-py to aioredis and give the pubsub system another try
        while True:
            record = self.redis.blpop("LOG_LIST", timeout=1)
            if record and "LOG_LIST" in record and record["LOG_LIST"]:
                yield jsonpickle.loads(record["LOG_LIST"])
            # time.sleep(0.005)

    def get_logs(self) -> List[logging.LogRecord]:
        return [
            jsonpickle.loads(record) for record in self.redis.lrange("LOG_LIST", 0, -1)
        ]



import os
import psutil
import graphio
import humanfriendly
from collections import OrderedDict
import time
import graphio
from typing import TYPE_CHECKING, Type, Union, Dict
from ..worker_base import Progress

if TYPE_CHECKING:
    from .manager import Manager


class Statistics:
    NodeSets_done_timeline: OrderedDict = None
    RelSets_done_timeline: OrderedDict = None
    graph_objects_loading_counts: OrderedDict = None
    management_ticks_timeline: OrderedDict = None

    def __init__(self, manager):
        self.manager: "Manager" = manager
        self.last_tick_sourcing_workers_done = []
        self.start_time = time.time()

        self.NodeSets_done_timeline = OrderedDict()
        self.RelSets_done_timeline = OrderedDict()
        self.graph_objects_loading_counts = OrderedDict()
        self.management_ticks_timeline = OrderedDict()

    def update_statistics(
        self, finished_sourcing_workers: list, finished_loading_workers: list
    ):
        current_time = time.time()
        # Save NodeSets/RelSets Workers timeline
        self.NodeSets_done_timeline[current_time] = [
            w for w in finished_loading_workers if w.set_meta.type == graphio.NodeSet
        ]

        self.RelSets_done_timeline[current_time] = [
            w
            for w in finished_loading_workers
            if w.set_meta.type == graphio.RelationshipSet
        ]

        # Save Labels/Rels amount loaded
        self.graph_objects_loading_counts[
            current_time
        ] = self.manager.cache.get_report_graphsets_loaded()

        # Save amount of management ticks
        self.management_ticks_timeline[current_time] = self.manager.management_ticks

    def get_management_ticks_per_n_sec(
        self, n: int = 1, timeframe: int = 20, object_type=None
    ) -> int:
        current_time = time.time()
        running_time = current_time - self.start_time
        if running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe

        relevant_data = OrderedDict()
        for timestamp, tick_count in self.management_ticks_timeline.items():
            if timestamp > timeframe_start:
                relevant_data[timestamp] = tick_count
        relevnt_tick_counts = list(relevant_data.values())
        if len(relevnt_tick_counts) < 2:
            tick_count = 0
        else:
            tick_count = (
                relevnt_tick_counts[len(relevant_data) - 1] - relevnt_tick_counts[0]
            )
        return (tick_count / (current_time - timeframe_start)) * n

    def get_graphobjects_count_loaded_per_n_sec(
        self, n: int = 1, timeframe: int = 20, object_type=None
    ) -> int:
        current_time = time.time()
        running_time = current_time - self.start_time
        if running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe
        relevant_data = OrderedDict()

        for timestamp, data in self.graph_objects_loading_counts.items():
            if timestamp > timeframe_start:
                relevant_data[timestamp] = data
        count_before_and_after = [0, 0]
        if len(relevant_data) == 0:
            return 0
        for index, data_position in enumerate([0, len(relevant_data) - 1]):
            count_all = 0
            report = list(relevant_data.values())[data_position]
            if object_type in [None, "Nodes"]:
                for graphobject_name, count in report["Nodes"].items():
                    count_all += count
            if object_type in [None, "Relations"]:
                for graphobject_name, count in report["Relations"].items():
                    count_all += count
            count_before_and_after[index] = count_all
        if len(relevant_data) < 2:
            count_before_and_after[0] = 0
        return ((count_before_and_after[1] - count_before_and_after[0]) / timeframe) * n

    def get_count_graphsets_loaded_per_n_sec(
        self,
        n: int = 1,
        timeframe: int = 20,
        graphset_type: Type[Union[graphio.NodeSet, graphio.RelationshipSet]] = None,
    ):
        current_time = time.time()
        running_time = current_time - self.start_time
        if running_time < timeframe:
            timeframe_start = self.start_time
        else:
            timeframe_start = current_time - timeframe

        workers = []
        for timestamp, workers_in_slot in self.NodeSets_done_timeline.items():
            if timestamp > timeframe_start:
                if graphset_type is None or graphset_type == graphio.NodeSet:
                    workers.extend(workers_in_slot)
                if graphset_type is None or graphset_type == graphio.RelationshipSet:
                    # this is a little bit hacky but we know for sure that the keys in self.RelSets_done_timeline and self.NodeSets_done_timeline are always in sync
                    workers.extend(self.RelSets_done_timeline[timestamp])
        return (len(workers) / (current_time - timeframe_start)) * n

    def get_count_running_sourcing_workers(self) -> int:
        return len(
            self.manager.manager_sourcing._get_workers(status=("initial", "running"))
        )

    def get_count_left_sourcing_workers(self) -> int:
        return len(self.manager.manager_sourcing.workers) - len(
            self.manager.manager_sourcing._get_workers(progress="complete")
        )

    def get_memory_consumed_by_loaders(self):
        mem_total_bytes = 0
        for worker in self.manager.manager_loading._get_workers(status="running"):
            process = psutil.Process(worker.pid)
            mem_total_bytes += process.memory_info()[0]
        return mem_total_bytes

    def get_memory_consumed_by_sourcing_workers(self):
        mem_total_bytes = 0
        for worker in self.manager.manager_sourcing._get_workers(status="running"):
            process = psutil.Process(worker.pid)
            mem_total_bytes += process.memory_info()[0]
        return mem_total_bytes

    def get_memory_consumed_by_manager(self):
        process = psutil.Process(os.getpid())
        return process.memory_info()[0]

    def is_sourcing_phase_done(self):
        return self.manager.manager_sourcing.is_done()

    def get_memory_available(self):
        return getattr(psutil.virtual_memory(), "available")

    def get_memory_total(self):
        return getattr(psutil.virtual_memory(), "total")

    def get_memory_used(self):
        return getattr(psutil.virtual_memory(), "used")

    def get_cache_storage_total(self) -> int:
        return self.manager.strategy.cache_storage_total

    def get_cache_storage_available(self) -> int:
        return self.get_cache_storage_total() - self.get_cache_storage_used()

    def get_cache_storage_used(self) -> int:
        return self.manager.cache.storage_used()

    def get_cache_storage_level(self) -> str:
        mem_full_perc = (
            self.get_cache_storage_used() * 100
        ) / self.get_cache_storage_total()
        if self.manager.strategy.cache_storage_clogged_limit <= mem_full_perc:
            return "RED"
        elif self.manager.strategy.cache_storage_warning_limit <= mem_full_perc:
            return "ORANGE"
        else:
            return "GREEN"

    def get_cache_storage_used_by_relSets(self):
        size = 0
        for rel_meta in self.manager.cache.list_SetsMeta(
            set_type=graphio.RelationshipSet
        ):
            size += rel_meta.total_size_bytes
        return size

    def get_cache_storage_used_by_nodeSets(self):
        size = 0
        for rel_meta in self.manager.cache.list_SetsMeta(set_type=graphio.NodeSet):
            size += rel_meta.total_size_bytes
        return size

    def get_report(self):
        report = {}

        report["General"] = {}

        gen_rep = report["General"]

        gen_rep["running_time_sec"] = {
            "val": time.time() - self.start_time,
            "desc": "Running time",
            "humanfriendly_transformer": humanfriendly.format_timespan,
        }

        gen_rep["management_tick_per_sec"] = {
            "val": self.get_management_ticks_per_n_sec(),
            "desc": "Management ticks per second",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        report["Sourcing"] = {}

        src_rep = report["Sourcing"]

        src_rep["cores_avail_no"] = {
            "val": self.manager.strategy.amount_sourcing_cores(),
            "desc": "Sourcing Workers amount of cores allocated",
        }
        src_rep["workers_total_no"] = {
            "val": len(self.manager.manager_sourcing.workers),
            "desc": "Sourcing Workers amount total",
        }

        src_rep["workers_fin_no"] = {
            "val": len(self.manager.manager_sourcing.finished_workers),
            "desc": "Sourcing Workers amount finished",
        }

        src_rep["workers_running_no"] = {
            "val": len(self.manager.manager_sourcing._get_workers(status="running")),
            "desc": "Sourcing Workers amount running",
        }

        report["Loading"] = {}

        load_rep = report["Loading"]

        load_rep["cores_avail_no"] = {
            "val": self.manager.strategy.amount_loading_cores(),
            "desc": "Loading Workers amount of cores allocated",
        }
        load_rep["workers_running_no"] = {
            "val": len(self.manager.manager_loading._get_workers(status="running")),
            "desc": "Loading Workers amount running",
        }
        load_rep["workers_fin_no"] = {
            "val": len(self.manager.manager_loading.finished_workers),
            "desc": "Loading Workers amount finished",
        }

        load_rep["workers_relsets_run_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    status=("running"),
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "Loading RelSets Workers amount running",
        }

        load_rep["workers_relsets_fin_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=Progress.COMPLETE,
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "Loading RelSets Workers amount finished",
        }

        load_rep["workers_relsets_wait_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=Progress.DRAIN_ORDERED,
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "RelSets Workers waiting for NodeSet drain",
        }

        load_rep["labels_blocked_for_rels"] = {
            "val": len(self.manager._get_blocked_labels()),
            "desc": "Amount of labels blocked for relation loading",
        }

        load_rep["relSet_per_sec"] = {
            "val": self.get_count_graphsets_loaded_per_n_sec(
                graphset_type=graphio.RelationshipSet
            ),
            "desc": "RelationSets loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        load_rep["rels_per_sec"] = {
            "val": self.get_graphobjects_count_loaded_per_n_sec(
                object_type="Relations"
            ),
            "desc": "Relations loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        load_rep["workers_relsets_run_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    status=("running"),
                    tag=self.manager.manager_loading.worker_tag_relsets,
                )
            ),
            "desc": "Loading RelSets Workers amount running",
        }

        load_rep["workers_NodeSets_run_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=(Progress.QUEUED, Progress.LOADING),
                    tag=self.manager.manager_loading.worker_tag_nodesets,
                )
            ),
            "desc": "Loading NodeSets Workers running or queded",
        }

        load_rep["workers_NodeSets_fin_no"] = {
            "val": len(
                self.manager.manager_loading._get_workers(
                    progress=Progress.COMPLETE,
                    tag=self.manager.manager_loading.worker_tag_nodesets,
                )
            ),
            "desc": "Loading NodeSets Workers amount finished",
        }

        load_rep["nodeSet_per_sec"] = {
            "val": self.get_count_graphsets_loaded_per_n_sec(
                graphset_type=graphio.NodeSet
            ),
            "desc": "NodeSets loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        load_rep["nodes_per_sec"] = {
            "val": self.get_graphobjects_count_loaded_per_n_sec(object_type="Nodes"),
            "desc": "Nodes loaded per sec",
            "humanfriendly_transformer": "{:.2f}/s".format,
        }

        report["Cache"] = {}

        cache_rep = report["Cache"]

        cache_rep["relSet_no"] = {
            "val": len(self.manager.cache.list_SetsMeta(graphio.RelationshipSet)),
            "desc": "RelationSets count in cache",
        }
        cache_rep["relSet_size"] = {
            "val": self.get_cache_storage_used_by_relSets(),
            "desc": "RelationSets size in cache",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        cache_rep["nodeSet_no"] = {
            "val": len(self.manager.cache.list_SetsMeta(graphio.NodeSet)),
            "desc": "NodeSets count in cache",
        }
        cache_rep["nodeSet_size"] = {
            "val": self.get_cache_storage_used_by_nodeSets(),
            "desc": "NodeSets size in cache",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        cache_rep["cache_storage_status"] = {
            "val": self.get_cache_storage_level(),
            "desc": "Cache storage health status",
        }

        cache_storage_level_details = f"{int(100 - ((self.get_cache_storage_used() * 100) / self.get_cache_storage_total()))}% Free ({humanfriendly.format_size(self.get_cache_storage_used())} used of {humanfriendly.format_size(self.get_cache_storage_total())})"

        cache_rep["cache_storage_status_details"] = {
            "val": cache_storage_level_details,
            "desc": "Cache storage status details",
        }

        report["Memory"] = {}

        mem_rep = report["Memory"]

        memory_status = f"{humanfriendly.format_size(self.get_memory_available())} available - {humanfriendly.format_size(self.get_memory_used())} used of {humanfriendly.format_size(self.get_memory_total())}"

        mem_rep["memory_status_details"] = {
            "val": memory_status,
            "desc": "Memory status for Manager and Workers processes",
        }

        mem_rep["memory_used_by_manager"] = {
            "val": self.get_memory_consumed_by_manager(),
            "desc": "Memory used by the manager process",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        mem_rep["memory_used_by_sourcing_workers"] = {
            "val": self.get_memory_consumed_by_sourcing_workers(),
            "desc": "Memory used by the sourcing processes",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        mem_rep["memory_used_by_loading_workers"] = {
            "val": self.get_memory_consumed_by_loaders(),
            "desc": "Memory used by the loading processes",
            "humanfriendly_transformer": humanfriendly.format_size,
        }

        return report

    def get_human_readable_report(self):

        report = self.get_report()

        report_str = "\n"
        for title, chapters in report.items():
            report_str += f"## {title}\n"
            longest_chapter_desc = max(
                [len(chapter["desc"]) for chapter in chapters.values()]
            )
            for key, chapter in chapters.items():
                if "humanfriendly_transformer" in chapter:
                    val = chapter["humanfriendly_transformer"](chapter["val"])
                else:
                    val = chapter["val"]
                row = "\t{:" + str(longest_chapter_desc) + "}: {}\n"
                report_str += row.format(chapter["desc"], val)

        return report_str



import multiprocessing
import logging


import time
from typing import List, Dict, Type, Set, Union
from ..worker_sourcing import WorkerSourcing
from ..worker_loading import WorkerLoading
from ..cache_backend import (
    CacheInterface,
    NodeSetMeta,
    RedisCache,
    SetsMetaBase,
)
from .manager_workers_loading import ManagerWorkersLoading
from .manager_workers_sourcing import ManagerWorkersSourcing
from .manager_strategies import StrategyBase
from .manager_statistics import Statistics
from ..cache_logger import LogRecordStreamHandler
import humanfriendly


log = logging.getLogger(__name__)


class Manager:
    # Redis - Default cache backend

    strategy_type: Type[StrategyBase] = StrategyBase

    def __init__(
        self,
        worker_sourcing_class: Type[WorkerSourcing],
        worker_parameters: List[Dict],
        cache_backend: Type[CacheInterface] = RedisCache,
        cache_backend_params: Dict = None,
        cache_size: str = "2000MB",
        debug: bool = False,
    ):
        """[summary]

        Args:
            worker_sourcing_class (Type[WorkerSourcing]): [description]
            worker_parameters (List[Dict]): [description]
            cache_backend (Type[CacheInterface], optional): [description]. Defaults to RedisCache.
            cache_backend_params (Dict, optional): [description]. Defaults to None.
            cache_size (str, optional): Max size of cache that should be available in humanfriendly ( https://pypi.org/project/humanfriendly/ ) size format . Defaults to "2000MB".
        """
        self.debug = debug

        self.worker_sourcing_class = worker_sourcing_class
        self.worker_parameters = worker_parameters
        self.worker_loading_class = WorkerLoading
        self.graph_params = {}
        self.cpu_count: int = multiprocessing.cpu_count()
        self.cache_backend = cache_backend
        self.cache_backend_params = cache_backend_params
        self.cache: CacheInterface = self.cache_backend(
            self.cache_backend_params, debug=self.debug
        )
        self.statistics = Statistics(self)
        self.strategy: StrategyBase = self.strategy_type(self)
        self.insert_action: str = "create"
        self.create_indexes = True
        self.create_unique_constraints = False
        self.worker_logging_reciever = LogRecordStreamHandler(self.cache)
        self.cache_size = humanfriendly.parse_size(cache_size)

        self.blocked_labels: Dict[str, Set[SetsMetaBase]] = {}
        self.update_statistics_per_sec = 4
        self.status_log_every_n_sec = 8

    def test_cache_backend_ready(self):
        if self.cache.test_connection():
            return True
        else:
            return False

    def is_cache_log_flushed(self):
        if len(self.cache.get_logs()) == 0:
            return True
        else:
            return False

    def merge(self, graph_params: Dict = {}):
        self.graph_params = graph_params
        self.insert_action = "merge"
        self._start()

    def create(self, graph_params: Dict = {}):
        self.graph_params = graph_params
        self.insert_action = "create"
        self._start()

    def _start(self):
        if not self.test_cache_backend_ready():
            raise Exception(
                f"Cache backend {type(self.cache_backend)} {self.cache_backend_params} not available!"
            )
        self.cache.clear_cache()
        self.manager_sourcing = ManagerWorkersSourcing(self, self.worker_sourcing_class)
        self.manager_loading = ManagerWorkersLoading(self, self.worker_loading_class)
        self.management_ticks = 0
        finished = False
        self.worker_logging_reciever.handle()
        log.info("##Management start##")
        last_status_log_time = time.time()
        last_statistics_update = time.time()
        while not finished:
            self.management_ticks += 1
            if time.time() - last_status_log_time > self.status_log_every_n_sec:
                self.log_status()
                last_status_log_time = time.time()
            if time.time() - last_statistics_update > self.update_statistics_per_sec:
                self.statistics.update_statistics(
                    self.manager_sourcing.just_finished_workers,
                    self.manager_loading.just_finished_workers,
                )
                self.manager_loading.just_finished_workers = []
                self.manager_sourcing.just_finished_workers = []
                last_statistics_update = time.time()

            log.debug(f"blocked_labels: {self.blocked_labels}")
            self.manager_sourcing.manage()
            self.manager_loading.manage()
            finished = (
                self.manager_sourcing.is_done() and self.manager_loading.is_done()
            )
        log.debug("##Management finish##")
        # Wait for all log records to be processed
        while not self.is_cache_log_flushed():
            time.sleep(0.1)
        self.statistics.update_statistics([], [])
        log.info(self.statistics.get_human_readable_report())

    def log_status(self):
        log.info(self.statistics.get_human_readable_report())

    def _block_loading_label(self, block_id: str, labels: Set[str]):
        log.debug(f"BLOCK {labels}")
        if block_id in self.blocked_labels:
            self.blocked_labels[block_id].update(labels)
        else:
            self.blocked_labels[block_id] = set(labels)

    def _release_loading_block_label(self, block_id: str):
        self.blocked_labels.pop(block_id, None)

    def _is_nodeset_loading_blocked(
        self, nodeset_meta: Union[SetsMetaBase, NodeSetMeta]
    ) -> bool:
        if isinstance(nodeset_meta, SetsMetaBase):
            nodeset_meta: NodeSetMeta = self.cache.get_NodeSetMeta(nodeset_meta)
        for label in nodeset_meta.labels:
            if label in self._get_blocked_labels():
                return True

    def _get_blocked_labels(self) -> tuple:
        # flatten all values in a list and remove duplicates
        return {
            label for label_list in self.blocked_labels.values() for label in label_list
        }

    def _manage_store(self):
        # if 90% of memory is occupied we dont allow any more sourcing workers to store data
        if self.strategy.get_cache_storage_level() == "RED":
            self.cache_backend.close_store()
            log.warning(
                "Cache backend is blocked for storing new NodeSets and RelationshipSets due to low memory. \n This will impact the overall perfomance dramaticly, but prevents the process from crashing. \n Try to work with a lower core count or get more memory."
            )
        elif not self.cache_backend.store_is_available():
            self.cache_backend.open_store()
            log.info(
                "Cache backend is open for storing NodeSets and RelationshipSets again"
            )



import operator
import os
import psutil
import graphio
import math
import humanfriendly
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .manager import Manager


class StrategyBase:
    cache_storage_warning_limit = 80
    cache_storage_clogged_limit = 90

    cache_storage_total = None
    # On large server with huge core counts, the sourcing workers can fly the cache quickly into the ceilling. thats why we create a default max amount of sourcing workers
    max_sourcing_workers = 10

    def __init__(self, manager: "Manager"):
        self.manager = manager
        if self.cache_storage_total is None:
            self.cache_storage_total = getattr(psutil.virtual_memory(), "total") * 0.80
        else:
            self.cache_storage_total = humanfriendly.parse_size(
                self.cache_storage_total
            )

    def amount_loading_cores(self) -> int:
        if (
            self.manager.statistics.is_sourcing_phase_done()
            or self.manager.statistics.get_cache_storage_used()
            >= self.manager.cache_size
            or self.manager.statistics.get_cache_storage_level() in ["ORANGE", "RED"]
        ):
            return self.manager.cpu_count
        c = round(self.manager.cpu_count - self.amount_sourcing_cores())
        if c == 0:
            c = 1
        return c * 2

    def amount_loading_nodes_cores(self) -> int:
        c = round(self.amount_loading_cores() * 0.9)
        if c == 0:
            c = 1
        # We can take more cores as available because loading processes are waiting a lot for the db and wont use any cpu time during that
        return math.ceil(c * 1.5)

    def amount_loading_rels_cores(self) -> int:
        # if we ended the sourcing phase (all sourcing workers finished) we first load all nodesets and after that all relsets can load without draining)
        if (
            self.manager.statistics.get_count_left_sourcing_workers() == 0
            and len(self.manager.cache.list_SetsMeta(graphio.NodeSet)) > 0
        ):
            return 0
        # if we are low on nodeSets in the cache we can start more RelSet loaders
        nodesets_cores_needed: int = len(
            self.manager.cache.list_SetsMeta(graphio.NodeSet)
            + self.manager.manager_loading._get_workers(
                status=("initial", "running"),
                tag=self.manager.manager_loading.worker_tag_nodesets,
            )
        )
        if nodesets_cores_needed < self.amount_loading_nodes_cores():
            return self.amount_loading_cores() - nodesets_cores_needed
            # else
        # By default we apply 50% of available loading cores to RelSets (or at least one Core)
        c = round(self.amount_loading_cores() * 0.5)
        if c == 0:
            c = 1
        return math.ceil(c)

    def amount_sourcing_cores(self) -> int:
        if self.manager.statistics.get_cache_storage_level() == "RED":
            return 0
        if (
            self.manager.statistics.get_cache_storage_used() >= self.manager.cache_size
            or self.manager.statistics.get_cache_storage_level() == "ORANGE"
        ):
            #  if we maxed out the allowed cache size, slow down sourcing new data into the cache and empty the cache by mainly allow cores to load data from the cache into the DB
            c = round(self.manager.cpu_count * 0.1)
        elif self.manager.statistics.get_count_running_sourcing_workers() >= round(
            self.manager.cpu_count * 0.6
        ):
            # provide 60% or max_sourcing_workers of cores if there are enough sourcing tasks
            c = round(self.manager.cpu_count * 0.6)
            if c > self.max_sourcing_workers:
                c = self.max_sourcing_workers
            # on CPUs with low core count we could round to zero cores. we want to have at least one core
            if c == 0:
                c = 1
        else:
            # if there are not many sourcing tasks left, we just provide enough cores for the leftovers
            c = self.manager.statistics.get_count_running_sourcing_workers()
        return c



import multiprocessing
from .cache_backend import CacheInterface
from typing import Dict, List
from linetimer import CodeTimer
from .cache_logger import CacheLoggerHandler
import logging
import uuid
from enum import Enum


class Progress(str, Enum):
    # str mixin, to be able to compare enums to strings. cool trick! see https://stackoverflow.com/a/63028809/12438690
    QUEUED = "queued"
    DRAIN_ORDERED = "drain"
    DRAIN_READY = "drained"
    LOADING = "load"
    COMPLETE = "complete"


class WorkerBase(multiprocessing.Process):
    cache_backend: CacheInterface = None
    cache_backend_params: Dict = None

    def __init__(self):
        super(WorkerBase, self).__init__()
        self.tags: List[str] = []
        self.params: Dict = {}
        self.timer: CodeTimer = CodeTimer(silent=True, unit="s")
        self.cache: CacheInterface = None
        self.id: str = uuid.uuid4().hex
        self.progress: Progress = Progress.QUEUED

    def run(self):
        pass

    def get_logger(self) -> logging.Logger:
        if self.cache:
            logging.basicConfig(level=logging.INFO, handlers=[])
            log = logging.getLogger(self.name)
            log.addHandler(CacheLoggerHandler(self.cache))
            return log
        else:
            raise KeyError(
                "WorkerBase self.cache is None. Initate cache backed first before call 'get_logger'"
            )


class WorkerPlaceHolder:
    """In some cases/systems joined and closed processes are still counted into ulimit. We need to delete the worker complettly to prevent this.
    But as we want to track closed workers and pull out some informations from them we create a copy from the workers interessing data before deleting.
    This class is the datawrapper/placeholder for these informations.
    """

    @classmethod
    def from_worker(cls, worker: WorkerBase):
        return cls(worker)

    def __init__(self, worker: WorkerBase):
        self.tags: List[str] = worker.tags
        self.params: Dict = worker.params
        self.timer = worker.timer
        self.id: str = worker.id
        self.progress = worker.progress
        if hasattr(worker, "set_meta"):
            self.set_meta = worker.set_meta





# Demo Script for neobulkmp
# This Script will insert some movies with a sample of actors with multiple processes
# Requirements:
# * neobulkmp installed `pip install neobulkmp`
# * A local neo4j DB as database `docker run --rm --network host -e NEO4J_AUTH=none neo4j`
# * A local redis DB as a loading cache `docker run --rm --network=host redis`

from neobulkmp import Manager, WorkerSourcing
import sys
import os
import logging
import graphio

logging.basicConfig(level=logging.INFO, format='%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s',
                    handlers=[logging.StreamHandler((sys.stdout))])


movies = []
movies.append({"title": "Matrix", "Actors": [
              "Keanu Reeves", "Carrie-Anne Moss", "Laurence Fishburne"]})
movies.append({"title": "Ghost in the Shell", "Actors": [
              "Atsuko Tanaka", "Akio Otsuka", "Iemasa Kayumi"]})
movies.append({"title": "Bill & Ted's Excellent Adventure", "Actors": [
              "Keanu Reeves", "Alex Winter", "George Carlin"]})
movies.append({"title": "Dogma", "Actors": [
              "Ben Affleck", "Jason Lee", "George Carlin"]})
movies.append({"title": "Almost Famous", "Actors": [
              "Patrick Fugit", "Jason Lee", "Frances McDormand"]})


def parsing_func(worker, title, Actors):
    # Create nodeset with movie
    movieset = graphio.NodeSet(["Movie"], merge_keys=["title"])
    movieset.add_node({"title": title})
    # Send Nodeset with movie to neobulkmp manager
    yield movieset
    # Create Actor Nodeset and connect actors to movie
    actorset = graphio.NodeSet(["Actor"], merge_keys=["name"])
    acts_in_movie = graphio.RelationshipSet(
        "ACTS_IN", ["Actor"], ["Movie"], ["name"], ["title"])
    for actor in Actors:
        actorset.add_node({"name": actor})
        acts_in_movie.add_relationship(
            {"name": actor}, {"title": title}, {})

    # Send actorset to neobulkmp manager
    yield actorset
    # Send actor to movie relation to neobulkmp manager
    yield acts_in_movie


# Mount parsing function on parsing workers
WorkerSourcing.sourcing_func = parsing_func

# Init manager, provide parsing worker and worker parameters
manager = Manager(worker_sourcing_class=WorkerSourcing,
                  worker_parameters=movies)
#manager.max_concurrently_worker_count = 1
manager.create_indexes = True
# manager.create_unique_constraints = True  # not yet implemented
manager.merge(graph_params={"host": "localhost"})



import logging
import graphio
from typing import List, Dict, Type, Set, TYPE_CHECKING
from ..worker_base import WorkerBase, Progress
from .manager_workers_base import ManagerWorkersBase


if TYPE_CHECKING:
    from .manager import Manager

log = logging.getLogger(__name__)


class ManagerWorkersSourcing(ManagerWorkersBase):
    def __init__(self, manager: "Manager", worker_class: Type[WorkerBase]):
        super(type(self), self).__init__(manager, worker_class)
        # create workers distribute task parameters per worker
        log.debug(f"self.parent.worker_parameters: {self.parent.worker_parameters}")
        self._init_workers(parameters=self.parent.worker_parameters)
        self.worker_count_total = len(self.workers)

    def is_done(self):
        # Mom, Are We There Yet?
        if self.worker_count_total == len(self.finished_workers + self.failed_workers):

            return True
        else:
            return False

    def manage(self):
        available_cores = self.parent.strategy.amount_sourcing_cores()
        # Collect all running sourcing workers
        waiting_workers = self._get_workers(progress=Progress.QUEUED)
        # Start next worker
        if (
            len(self._get_workers(progress=Progress.LOADING)) < available_cores
            and len(waiting_workers) > 0
        ):
            next_worker = waiting_workers.pop(0)
            log.debug(f"CALL START ON SOURCING {next_worker} - {next_worker.params}")
            next_worker.progress = Progress.LOADING
            next_worker.start()
            print("ME STARTING")
            next_worker.timer.__enter__()

        self._finish_workers()



# Demo Script for neobulkmp
# This Script will insert some movies with a sample of actors with multiple processes
# Requirements:
# * neobulkmp installed `pip install neobulkmp`
# * A local neo4j DB as database `docker run --rm --network host -e NEO4J_AUTH=none neo4j`
# * A local redis DB as a loading cache `docker run --rm --network=host redis`

from neobulkmp import Manager, WorkerSourcing
import sys
import os
import logging
import graphio

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s",
    handlers=[logging.StreamHandler((sys.stdout))],
)


movies = []
movies.append(
    {
        "title": "Matrix",
        "Actors": ["Keanu Reeves", "Carrie-Anne Moss", "Laurence Fishburne"],
    }
)
movies.append(
    {
        "title": "Ghost in the Shell",
        "Actors": ["Atsuko Tanaka", "Akio Otsuka", "Iemasa Kayumi"],
    }
)
movies.append(
    {
        "title": "Bill & Ted's Excellent Adventure",
        "Actors": ["Keanu Reeves", "Alex Winter", "George Carlin"],
    }
)
movies.append(
    {"title": "Dogma", "Actors": ["Ben Affleck", "Jason Lee", "George Carlin"]}
)
movies.append(
    {
        "title": "Almost Famous",
        "Actors": ["Patrick Fugit", "Jason Lee", "Frances McDormand"],
    }
)


class MyCustomWorker(WorkerSourcing):
    def sourcing_func(worker, title, Actors):
        # Create nodeset with movie
        movieset = graphio.NodeSet(["Movie"], merge_keys=["title"])
        movieset.add_node({"title": title})
        # Send Nodeset with movie to neobulkmp manager
        yield movieset
        # Create Actor Nodeset and connect actors to movie
        actorset = graphio.NodeSet(["Actor"], merge_keys=["name"])
        acts_in_movie = graphio.RelationshipSet(
            "ACTS_IN", ["Actor"], ["Movie"], ["name"], ["title"]
        )
        for actor in Actors:
            actorset.add_node({"name": actor})
            acts_in_movie.add_relationship({"name": actor}, {"title": title}, {})

        # Send actorset to neobulkmp manager
        yield actorset
        # Send actor to movie relation to neobulkmp manager
        yield acts_in_movie


# Init manager, provide parsing worker and worker parameters
manager = Manager(worker_sourcing_class=MyCustomWorker, worker_parameters=movies)
# manager.max_concurrently_worker_count = 1
manager.create_indexes = True
# manager.create_unique_constraints = True  # not yet implemented
manager.merge(graph_params={"host": "localhost"})



import graphio
from typing import Callable, Any, Union, Generator, Dict
from .worker_base import WorkerBase
from logging import Logger
from .cache_backend import CacheInterface
import time


class WorkerSourcing(WorkerBase):
    sourcing_func: Callable[
        [Any], Generator[Union[graphio.NodeSet, graphio.RelationshipSet], None, None]
    ] = None

    def __init__(self, name, **kwargs):
        super().__init__()
        self.name = name
        self.worker_parameter = kwargs

    def run(self):
        self.cache: CacheInterface = self.cache_backend(self.cache_backend_params)

        log: Logger = self.get_logger()
        log.debug(f"START {self.name}")
        if not isinstance(self.worker_parameter, dict):
            raise ValueError("worker_parameter not a dict:", self.worker_parameter)
        for set in self.sourcing_func(**self.worker_parameter):
            data_stored = False
            while not data_stored:
                if self.cache.store_is_available():
                    if isinstance(set, graphio.NodeSet):
                        log.debug(f"Recieved graphio.NodeSet '{set.labels}'")
                        self.cache.store_NodeSet(set)
                    elif isinstance(set, graphio.RelationshipSet):
                        log.debug(f"Recieved graphio.RelationshipSet '{set.rel_type}'")
                        self.cache.store_RelSet(set)
                    else:
                        raise ValueError(
                            f"Worker '{self.name}' yielded wrong type of data. Expected 'graphio.NodeSet' or 'graphio.RelationshipSet' got '{type(set)}' "
                        )
                    data_stored = True
                else:
                    # wait 10 second for store to open again and try again
                    time.sleep(10)



from setuptools import setup

setup(
    name="neobulkmp",
    description="Tool for multiprocessed dataloading into a neo4j db, without node blocking",
    url="",
    author="Tim Bleimehl",
    author_email="tim.bleimehl@helmholtz-muenchen.de",
    license="MIT",
    packages=["neobulkmp"],
    install_requires=[
        "py2neo",
        "graphio",
        "linetimer",
        "redis",
        "jsonpickle",
        "psutil",
        "humanfriendly",
    ],
    extras_require={"test": ["DZDutils"]},
    python_requires=">=3.6",
    zip_safe=False,
    include_package_data=True,
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        # "local_scheme": "node-and-timestamp"
        "local_scheme": "no-local-version",
        "write_to": "version.py",
    },
    setup_requires=["setuptools_scm"],
)
